/*
 *  Copyright (C) 2005-2006 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 * The exception module defines all system-level exceptions and provides a
 * mechanism to alter system-lever error handling.
 *
 * Design Issues:
 *
 * The Exception base class is required to live in the global object module,
 * as it is used by the system for exception propogation.
 *
 * Future Directions:
 *
 * It may be useful to provide a means of determining whether an exception
 * is currently in flight.  This would require cooperation with the thread
 * module to implement.
 */
module std.exception;

private
{
    import std.c.stdio;
    import std.c.stdlib;
}

private
{
    alias void function( char[] file, uint line ) assertHandlerType;
    alias void function() terminateHandlerType;

    assertHandlerType assertHandler = null;
}

/**
 * Thrown on an array bounds error.
 */
class ArrayBoundsException : Exception
{
    this( char[] file, size_t line )
    {
        super( "array index out of bounds", file, line );
    }
}

/**
 * Thrown on an assert error.
 */
class AssertException : Exception
{
    this( char[] file, size_t line )
    {
        super( "assertion failure", file, line );
    }
}

/**
 * Thrown on an out of memory error.
 */
class OutOfMemoryException : Exception
{
    this( char[] file, size_t line )
    {
        super( "memory allocation failed", file, line );
    }
}

/**
 * Thrown on a switch error.
 */
class SwitchException : Exception
{
    this( char[] file, size_t line )
    {
        super( "no appropriate switch clause found", file, line );
    }
}

/**
 * Thrown on a unicode conversion error.
 */
class UnicodeException : Exception
{
    size_t idx;

    this( char[] msg, size_t idx )
    {
        super( msg );
        this.idx = idx;
    }
}

/**
 * Overrides the default assert hander with a user-supplied version.
 *
 * Params:
 *  h = The new assert handler.
 */
void setAssertHandler( assertHandlerType h )
{
    assertHandler = h;
}

/**
 * A callback for assert errors in D.  The user-supplied assert handler will
 * be called if one has been supplied, otherwise an AssertException will be
 * thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onAssert( char[] file, uint line )
{
    if( assertHandler )
        assertHandler( file, line );
    else
        throw new AssertException( file, line );
}

/**
 * A callback for out of memory errors in D.  An OutOfMemoryException will be
 * thrown.
 */
extern (C) void onOutOfMemory()
{
    throw cast(OutOfMemoryException) cast(void*) OutOfMemoryException.classinfo.init;
}

/**
 * A callback for array bounds errors in D.  An ArrayBoundsException will be
 * thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onArrayBoundsError( char[] file, size_t line )
{
    throw new ArrayBoundsException( file, line );
}

/**
 * A callback for switch errors in D.  A SwitchException will be thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onSwitchError( char[] file, size_t line )
{
    throw new SwitchException( file, line );
}

/**
 * A callback for unicode errors in D.  A UnicodeException will be thrown.
 *
 * Params:
 *  msg = Information about the error.
 *  idx = String index where this error was detected.
 */
extern (C) void onUnicodeError( char[] msg, size_t idx )
{
    throw new UnicodeException( msg, idx );
}