/*
 *  Copyright (C) 2005-2006 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 * The string module provides string manipulation routines in a manner that
 * balances performance and flexibility.
 *
 * Design Issues:
 *
 * While these functions will operate with wide character types, little attempt
 * has yet been made to provide any explicit Unicode support.  So case-insensitive
 * routines are not available and equality is done via binary comparisons.
 *
 * Future Directions:
 *
 * This module is the first attempt at developing function interface guidelines
 * for Ares.  All functions will likely be templates and will be designed with
 * implicit tempalte instantiation and template overloading as required features.
 * The goal here is to provide a useful set of string routines for everyday use
 * and to use this experience for developing more generalized algorithms later on.
 */
module std.string;


private import std.c.string;


private template isCharType( Elem )
{
    const bool isCharType = is( Elem == char )  ||
                            is( Elem == wchar ) ||
                            is( Elem == dchar );
}


/**
 *
 */
 /+
template find( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     */
    size_t find( Elem[] str, Elem chr )
    {
        for( size_t pos = 0; pos < str.length; ++pos )
        {
            if( str[pos] == chr )
                return pos;
        }
        return size_t.max;
    }
}


unittest
{
    assert( find!(char)( "", 'a' ) == size_t.max );
    assert( find!(char)( "abc", 'a' ) == 0 );
    assert( find!(char)( "abc", 'b' ) == 1 );
    assert( find!(char)( "abc", 'c' ) == 2 );
    assert( find!(char)( "abc", 'd' ) == size_t.max );
}
+/


/**
 *
 */
template find( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     */
    size_t find( Elem[] str, Elem[] pat )
    {
        if( str.length == 0 ||
            pat.length == 0 ||
            str.length < pat.length )
        {
            return size_t.max;
        }

        size_t end = str.length - pat.length + 1;

        for( size_t pos = 0; pos < end; ++pos )
        {
            if( str[pos] == pat[0] )
            {
                size_t mat = 0;

                do
                {
                    if( ++mat >= pat.length )
                        return pos - pat.length + 1;
                    if( ++pos >= str.length )
                        return size_t.max;
                } while( str[pos] == pat[mat] );
                pos -= mat;
            }
        }
        return size_t.max;
    }
}


unittest
{
    // null parameters
    assert( find!(char)( "", "" ) == size_t.max );
    assert( find!(char)( "a", "" ) == size_t.max );
    assert( find!(char)( "", "a" ) == size_t.max );

    // exact match
    assert( find!(char)( "abc", "abc" ) == 0 );

    // simple substring match
    assert( find!(char)( "abc", "a" ) == 0 );
    assert( find!(char)( "abca", "a" ) == 0 );
    assert( find!(char)( "abc", "b" ) == 1 );
    assert( find!(char)( "abc", "c" ) == 2 );
    assert( find!(char)( "abc", "d" ) == size_t.max );

    // multi-char substring match
    assert( find!(char)( "abc", "ab" ) == 0 );
    assert( find!(char)( "abcab", "ab" ) == 0 );
    assert( find!(char)( "abc", "bc" ) == 1 );
    assert( find!(char)( "abc", "ac" ) == size_t.max );
    assert( find!(char)( "abrabracadabra", "abracadabra" ) == 3 );
}


/**
 *
 */
 /+
template rfind( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     */
    size_t rfind( Elem[] str, Elem chr )
    {
        if( str.length == 0 )
            return size_t.max;

        size_t pos = str.length;

        do
        {
            if( str[--pos] == chr )
                return pos;
        } while( pos > 0 );
        return size_t.max;
    }
}


unittest
{
    assert( rfind!(char)( "", 'a' ) == size_t.max );
    assert( rfind!(char)( "abc", 'a' ) == 0 );
    assert( rfind!(char)( "abc", 'b' ) == 1 );
    assert( rfind!(char)( "abc", 'c' ) == 2 );
    assert( rfind!(char)( "abc", 'd' ) == size_t.max );
}
+/


/**
 *
 */
template rfind( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     */
    size_t rfind( Elem[] str, Elem[] pat )
    {
        if( str.length == 0 ||
            pat.length == 0 ||
            str.length < pat.length )
        {
            return size_t.max;
        }

        size_t pos = str.length - pat.length + 1;
        do
        {
            if( str[--pos] == pat[0] )
            {
                size_t mat = 0;
                do
                {
                    if( ++mat >= pat.length )
                        return pos - pat.length + 1;
                    if( ++pos >= str.length )
                        return size_t.max;
                } while( str[pos] == pat[mat] );
                pos -= mat;
            }
        } while( pos > 0 );
        return size_t.max;
    }
}


unittest
{
    // null parameters
    assert( rfind!(char)( "", "" ) == size_t.max );
    assert( rfind!(char)( "a", "" ) == size_t.max );
    assert( rfind!(char)( "", "a" ) == size_t.max );

    // exact match
    assert( rfind!(char)( "abc", "abc" ) == 0 );

    // simple substring match
    assert( rfind!(char)( "abc", "a" ) == 0 );
    assert( rfind!(char)( "abca", "a" ) == 3 );
    assert( rfind!(char)( "abc", "b" ) == 1 );
    assert( rfind!(char)( "abc", "c" ) == 2 );
    assert( rfind!(char)( "abc", "d" ) == size_t.max );

    // multi-char substring match
    assert( rfind!(char)( "abc", "ab" ) == 0 );
    assert( rfind!(char)( "abcab", "ab" ) == 3 );
    assert( rfind!(char)( "abc", "bc" ) == 1 );
    assert( rfind!(char)( "abc", "ac" ) == size_t.max );
    assert( rfind!(char)( "abracadabrabra", "abracadabra" ) == 0 );
}


/**
 *
 */
template findAll( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     */
    size_t[] findAll( Elem[] str, Elem chr )
    {
        size_t[] res;

        for( size_t pos = 0; pos < str.length; ++pos )
        {
            if( str[pos] == chr )
                res ~= pos;
        }
        return res;
    }
}


unittest
{
    size_t[] res;
    res = findAll!(char)( "abracadabra", 'x' );
    assert( res.length == 0 );
    res = findAll!(char)( "abracadabra", 'a' );
    assert( res.length == 5 );
    assert( res[0] == 0 );
    assert( res[1] == 3 );
    assert( res[2] == 5 );
    assert( res[3] == 7 );
    assert( res[4] == 10 );
}


/**
 *
 */
template split( Elem )
{
    static assert( isCharType!(Elem) );

    /**
     *
     * Author:
     *  Walter Bright, Digital Mars, www.digitalmars.com
     * Copyright:
     *  Public Domain
     */
    Elem[][] split( Elem[] str )
    {
        size_t      beg    = 0;
        size_t      pos    = 0;
        bool        inword = false;
        Elem[][]    words;

        for( pos = 0; pos < str.length; ++pos )
        {
        	switch( str[pos] )
        	{
            case ' ':
            case '\t':
            case '\f':
            case '\r':
            case '\n':
            case '\v':
        		if( inword )
        		{
        		    words ~= str[beg .. pos];
        		    inword = 0;
        		}
        		break;
            default:
        		if( !inword )
        		{   beg = pos;
        		    inword = 1;
        		}
        		break;
        	}
        }
        if( inword )
    	    words ~= str[beg .. pos];
        return words;
    }
}


unittest
{
    char[]   str   = " peter paul\tjerry ";
    char[][] words = split( str );

    assert( words.length == 3 );
    assert( words[0] == "peter" );
    assert( words[1] == "paul" );
    assert( words[2] == "jerry" );
}