/*
 *  Copyright (C) 2005-2006 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 * The thread module provides support for thread creation and management.
 *
 * Design Issues:
 *
 * One design goal of Ares is to avoid forcing the use of a particular
 * programming style, so long as allowing such flexibility does not
 * compromise the overall API design.  This goal was realized here by
 * allowing threads to be created in the familiar C style (ie. by
 * composition), or by derivation, similar to the Java style.  Composition
 * is further supported by virtue of the thread local storage facility,
 * which allows for thread local data to be stored by the main thread as
 * well as by user threads.
 *
 * Future Directions:
 *
 * Support for lightwewight user threads is a long-term consideration,
 * though the design of this module is largely settled for now.
 */
module std.thread;


/**
 * All exceptions thrown from this module derive from this class.
 */
class ThreadError : Exception
{
    this( char[] msg )
    {
        super( msg );
    }
}


// This should be true for most architectures.
version = StackGrowsDown;


private
{
    import std.c.string; // for memset, memcpy
}


version( Win32 )
{
    private
    {
        import sys.windows.c.process;
        import sys.windows.c.windows;


        //
        // entry point for Windows threads
        //
        extern (Windows) uint threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        assert( obj );
	        scope(exit) Thread.remove( obj );

	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

            obj.m_bstack = getStackBottom();
            obj.m_tstack = obj.m_bstack;
            TlsSetValue( Thread.sm_this, obj );

	        try
	        {
	            obj.run();
	        }
	        catch
	        {
	            // error should really print to stderr
	        }
	        return 0;
        }


        //
        // copy of the same-named function in phobos.std.thread,
        // it uses the Windows naming convention to be consistent
        // with GetCurrentThreadId
        //
        HANDLE GetCurrentThreadHandle()
        {
            const uint DUPLICATE_SAME_ACCESS = 0x00000002;

    	    HANDLE curr = GetCurrentThread(),
    	           proc = GetCurrentProcess(),
    	           hndl;

    	    DuplicateHandle( proc, curr, proc, &hndl, 0, TRUE, DUPLICATE_SAME_ACCESS );
    	    return hndl;
        }


        void* getStackBottom()
        {
            asm
            {
                naked;
                mov	EAX, FS:4;
                ret;
            }
        }


        void* getStackTop()
        {
            asm
            {
        	    naked;
                mov EAX, ESP;
                ret;
            }
        }
    }
}
else version( Posix )
{
    private
    {
        import std.c.posix.semaphore;
        import std.c.posix.pthread;
        import std.c.posix.signal;
        import std.c.posix.unistd;
        import std.c.posix.time;

        version( darwin )
            import sys.darwin.c.darwin;
        else
            import sys.linux.c.linux;


        //
        // entry point for POSIX threads
        //
        extern (C) void* threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        assert( obj );
	        scope(exit) Thread.remove( obj );

	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

	        static extern (C) void cleanupHandler( void* arg )
	        {
                Thread  obj = cast(Thread) arg;
                assert( obj );

                obj.m_isRunning = false;
	        }

            obj.m_bstack = getStackBottom();
	        obj.m_tstack = obj.m_bstack;
	        pthread_setspecific( obj.sm_this, obj );

	        //
	        // NOTE: isRunning must return true for this thread before it is added
	        //       to the global thread list or it may be mistaken as completed
	        //       by a badly timed GC run.
	        //
	        obj.m_isRunning = true;
	        pthread_cleanup cleanup;
	        cleanup.push( &cleanupHandler, obj );

	        try
	        {
	            obj.run();
	        }
	        catch
	        {
	            // error should really print to stderr
	        }

            //
            // NOTE: The cleanup handler is not automatically run if a thread exits
            //       normally but we can call it here by popping with a nonzero value.
            //
	        cleanup.pop( 1 );
	        return null;
        }


        extern (C) void suspendHandler( int sig )
        in
        {
            assert( sig == SIGUSR1 );
        }
        body
        {
            Thread  obj = Thread.getThis();
            assert( obj );

            obj.m_tstack = getStackTop();

            version( X86 )
            {
                asm
                {
                    pushad;
                }
            }
            // FIXME: maybe that function? (how to pop again...)
            //else version( PPC )
            //{
            //    __builtin_unwind_init();
            //}
            else
            {
                static assert( false );
            }

            assert( getStackTop() < obj.m_tstack );
            obj.m_freg = obj.m_tstack - 1;
            obj.m_lreg = getStackTop();

            sigset_t signal_set;
            sigdelset( &signal_set, SIGUSR2 );
            sem_post( &obj.m_suspended );
            sigsuspend( &signal_set );

            obj.m_lreg = null;
            obj.m_freg = null;

            version( X86 )
            {
	            asm
	            {
    	            popad;
	            }
	        }
            // FIXME: maybe that function? (how to pop again...)
            //else version( PPC )
            //{
            //    __builtin_unwind_init();
            //}
	        else
	        {
	            static assert( false );
	        }
	        obj.m_tstack = obj.m_bstack;
        }


        extern (C) void resumeHandler( int sig )
        in
        {
            assert( sig == SIGUSR2 );
        }
        body
        {

        }


        //
        // NOTE: this may not work on all versions of linux,
        //       but apparently the Windows implementation
        //       does not work in Linux (according to DMD revs)
        //
        void* getStackBottom()
        {
            version( darwin )
                return _d_gcc_query_stack_origin();
            else
                return __libc_stack_end;
        }


        void* getStackTop()
        {
            version( X86 )
            {
            	asm
            	{   naked;
            	    mov EAX, ESP;
            	    ret;
            	}
            }
            else
            {
                static assert( false );
            }
        }
    }
}
else
{
    // NOTE: This is the only place threading versions are checked.  If a new
    //       version is added, the module code will need to be searched for
    //       places where version-specific code may be required.  This can be
    //       easily accomlished by searching for 'Windows' or 'Posix'.

    // Unknown threading implementation
    static assert( false );
}


////////////////////////////////////////////////////////////////////////////////
// Thread
////////////////////////////////////////////////////////////////////////////////


/**
 * This class encapsulates all threading functionality for the D
 * programming language.  As thread manipulation is a required facility
 * for garbage collection, all user threads should derive from this
 * class, and instances of this class should never be explicitly deleted.
 * A new thread may be created using either derivation or composition, as
 * in the following example.
 *
 * Example:
 * ----------------------------------------------------------------------
 *
 * class DerivedThread : Thread
 * {
 * protected:
 *     void run()
 *     {
 *         printf( "Derived thread running.\n" );
 *     }
 * }
 *
 * void threadFunc()
 * {
 *     printf( "Ccomposed thread running!\n" );
 * }
 *
 * // create instances of each type
 * Thread derived = new DerivedThread();
 * Thread composed = new Thread( &threadFunc );
 *
 * // start both threads
 * derived.start();
 * composed.start();
 *
 * // wait for the threads to complete
 * derived.join();
 * composed.join();
 *
 * ----------------------------------------------------------------------
 */
class Thread
{
    ////////////////////////////////////////////////////////////////////////////
    // Initialization
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Initializes a thread object which has no associated executable
     * function.
     */
    this()
    {
        m_call = Call.NO;

        init();
    }


    /**
     * Initialized a thread object which is associated with a static
     * D function.
     *
     * Params:
     *  fn = The thread function.
     */
    this( void function() fn )
    {
        m_fn   = fn;
        m_call = Call.FN;

        init();
    }


    /**
     * Initializes a thread object which is associated with a dynamic
     * D function.
     *
     * Params:
     *  dg = The thread function.
     */
    this( void delegate() dg )
    {
        m_dg   = dg;
        m_call = Call.DG;

        init();
    }


    //
    // NOTE: This function should be called by all ctors, and only by ctors.
    //
    private void init()
    {
        version( Posix )
        {
            if( sem_init( &m_suspended, 0, 0 ) != 0 )
                throw new ThreadError( "Error initializing thread data" );
        }
    }


    /**
     * Performs routine cleanup for this object.
     *
     * In:
     *  assert( !isRunning() );
     */
    ~this()
    in
    {
        assert( !isRunning() );
    }
    body
    {
        if( m_addr )
        {
            remove( this );
        }

        version( Win32 )
        {
            CloseHandle( m_hndl );
        }
        else version( Posix )
        {
            sem_destroy( &m_suspended );
            pthread_detach( m_addr );
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // General Actions
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Starts the thread with run() as the target method.  The default
     * behavior of this is to run the function or delegate passed upon
     * construction.
     *
     * In:
     *  This routine may only be called once per thread instance.
     *
     * Throws:
     *  ThreadError if the thread fails to start.
     */
    final void start()
    in
    {
        assert( !m_addr );
    }
    body
    {
        synchronized( slock )
        {
            version( Win32 )
            {
                m_hndl = cast(HANDLE) _beginthreadex( null, 0, &threadFunc, this, 0, &m_addr );
                if( cast(size_t) m_hndl == 0 )
                    throw new ThreadError( "Error creating thread" );
            }
            else version( Posix )
            {
                if( pthread_create( &m_addr, null, &threadFunc, this ) != 0 )
                    throw new ThreadError( "Error creating thread" );
            }
            multiThreadedFlag = true;
            add( this );
        }
    }


    /**
     * Waits for this thread to complete.
     *
     * Throws:
     *  ThreadError if the operation fails.
     */
    final void join()
    {
        version( Win32 )
        {
            if( WaitForSingleObject( m_hndl, INFINITE ) != WAIT_OBJECT_0 )
                throw new ThreadError( "Unable to join thread" );
        }
        else version( Posix )
        {
            if( pthread_join( m_addr, null ) != 0 )
                throw new ThreadError( "Unable to join thread" );
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // General Properties
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Gets the user-readable label for this thread.
     *
     * Returns:
     *  The name of this thread.
     */
    final char[] name()
    {
        synchronized( this )
        {
            return m_name;
        }
    }


    /**
     * Sets the user-readable label for this thread.
     *
     * Params:
     *  n = The new name for this thread.
     */
    final void name( char[] n )
    {
        synchronized( this )
        {
            m_name = n.dup;
        }
    }


    /**
     * Tests whether this thread is running.  This function should be
     * callable from anywhere within the application without a risk of
     * deadlock.
     *
     * Returns:
     *  true if the thread is running, false if not.
     */
    final bool isRunning()
    {
        if( !m_addr )
        {
            return false;
        }

        version( Win32 )
        {
            uint ecode = 0;
            GetExitCodeThread( m_hndl, &ecode );
            return ecode == STILL_ACTIVE;
        }
        else version( Posix )
        {
            // NOTE: It should be safe to access this value without
            //       memory barriers because
            return m_isRunning;
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Debug Actions
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Suspends thread execution and loads stack and register data for use
     * by the suspending thread.  Registers should be stored in the order
     * they would typically appear when pushed onto the execution stack.
     * This function is reentrant, but resume must be called once for each
     * call to suspend.
     *
     * Throws:
     *  ThreadError if the thread is not currently running or if the thread
     *  can not be suspended.
     */
    final void suspend()
    {
        // NOTE: this operation uses the global shared lock for the following reasons:
        //         - typically, this function is called by suspendAll during a GC run,
        //           and it would be prohibitively expensive to have a separate synch
        //           block for every thread that needs to be suspended
        //         - implementing a correct lockless per-thread synchronization mechanism
        //           is too complicated to be worthwhile
        //         - this prevents the unlikely chance that a thread may try to suspend
        //           another thread that is beginning a GC run, thus deadlocking the app
        //       this implementation also relies on synch blocks being reentrant, as it
        //       is quite likely that a GC run will first lock the thread list and then
        //       attempt to pause each thread
        //
        // NOTE: it might be useful to use DBC to ensure that the suspending thread is
        //       the only thread that accesses this thread's stack pointers and such.
        synchronized( slock )
        {
            if( m_scount > 0 )
            {
                ++m_scount;
                return;
            }

            version( Win32 )
            {
                // NOTE: atempting to suspend a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && SuspendThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to suspend thread" );
                ++m_scount;

    	        CONTEXT context;
    	        context.ContextFlags = CONTEXT_INTEGER | CONTEXT_CONTROL;

    	        if( !GetThreadContext( m_hndl, &context ) )
    	            throw new ThreadError( "Unable to load thread context" );
                m_tstack = cast(void*) context.Esp;
                // edi,esi,ebp,esp,ebx,edx,ecx,eax
                m_reg[0] = context.Edi;
                m_reg[1] = context.Esi;
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                m_reg[4] = context.Ebx;
                m_reg[5] = context.Edx;
                m_reg[6] = context.Ecx;
                m_reg[7] = context.Eax;
                /+
                memcpy( &m_reg[0], &context.Edi, uint.sizeof * 2 );
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                memcpy( &m_reg[4], &context.Ebx, uint.sizeof * 4 );
                +/
            }
            else version( Posix )
            {
                if( m_addr == pthread_self() )
                {
                    ++m_scount;

                    m_tstack = getStackTop();
    	            asm
    	            {
    	                pushad;
    	            }
    	            assert( getStackTop() < m_tstack );
    	            m_freg = m_tstack - 1;
    	            m_lreg = getStackTop();
                }
                else
                {
                    // NOTE: atempting to suspend a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR1 ) != 0 )
                        throw new ThreadError( "Unable to suspend thread" );
                    sem_wait( &m_suspended );
                    ++m_scount;
                }
            }
        }
    }


    /**
     * Resumes thread execution and frees stack and register data.  This
     * function is reentrant and must be called once for every call to
     * suspend before the thread is resumed.
     *
     * Throws:
     *  ThreadError if the thread is not currently running or if the thread
     *  can not be resumed.
     */
    final void resume()
    {
        // NOTE: see comments in suspend for why this shared lock is used
        synchronized( slock )
        {
            if( m_scount < 1 )
                return;
            if( m_scount > 1 )
            {
                --m_scount;
                return;
            }

            version( Win32 )
            {
                // NOTE: atempting to resume a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && ResumeThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to resume thread" );
                --m_scount;

                m_tstack = m_bstack;
                memset( &m_reg[0], 0, uint.sizeof * m_reg.length );
            }
            else version( Posix )
            {
                if( m_addr == pthread_self() )
                {
                    // pop registers off the stack before resuming
	                asm
	                {
	                    popad;
	                }
                }
                else
                {
                    // NOTE: atempting to resume a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR2 ) != 0 )
                        throw new ThreadError( "Unable to resume thread" );
                }
                --m_scount;

                m_tstack = m_bstack;
                m_freg   = null;
                m_lreg   = null;
            }
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Debug Properties
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Should be preceded by a call to suspend.
     *
     * Returns:
     *  A pointer to the logical top of the execution stack for this thread.
     *  Note that on some operating systems the address may be less than
     *  the address returned by stackBottom.
     */
    final void* stackTop()
    {
        return m_tstack;
    }


    /**
     * Should be preceded by a call to suspend.
     *
     * Returns:
     *  A pointer to the logical bottom of the execution stack for this
     *  thread.  Note that on some operating systems the address may be
     *  greater than the address returned by stackTop.
     */
    final void* stackBottom()
    {
        return m_bstack;
    }


    /**
     * Should be preceded by a call to suspend.
     *
     * Returns:
     *  A pointer to the beginning of an inclusive range of registers
     *  stored by suspend.  For example, if a system has registers X, Y, Z,
     *  and they are typically pushed onto the stack in that order, then
     *  this function will point to the value of X on systems where
     *  the stack grows up, or Z on systems where the stack grows down.
     */
    final void* regFirst()
    {
        version( Win32 )
        {
            return &m_reg[0];
        }
        else version( Posix )
        {
            return m_freg;
        }
    }


    /**
     * Should be preceded by a call to suspend.
     *
     * Returns:
     *  A pointer to the end of an inclusive range of registers stored by
     *  suspend.  For example, if a system has registers X, Y, Z, and they
     *  are typically pushed onto the stack in that order, then this
     *  function will point to the value of Z on systems where the stack
     *  grows up, or X on systems where the stack grows down.
     */
    final void* regLast()
    {
        version( Win32 )
        {
            return &m_reg[length-1];
        }
        else version( Posix )
        {
            return m_lreg;
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Static Initialization
    ////////////////////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////////////////////
    // Actions on Calling Thread
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Suspends the calling thread for at least the supplied time.
     *
     * Params:
     *  milliseconds = The minimum duration the calling thread should be
     *                 suspended.
     */
    static void sleep( uint milliseconds )
    {
        version( Win32 )
        {
            Sleep( milliseconds );
        }
        else version( Posix )
        {
            if( uint.max / 1000 < milliseconds )
                sleep( milliseconds / 1000 );
            milliseconds %= 1000;
            if( milliseconds > 0 )
                usleep( milliseconds * 1000 );
        }
    }


    /**
     * Forces a context switch to occur away from the calling thread.
     */
    static void yield()
    {
        version( Win32 )
        {
            // NOTE: Sleep(1) is necessary because Sleep(0) does not give
            //       lower priority threads any timeslice, so looping on
            //       Sleep(0) could be resource-intensive in some cases.
            Sleep( 1 );
        }
        else version( Posix )
        {
            sched_yield();
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Thread Accessors
    ////////////////////////////////////////////////////////////////////////////


    /**
     * The calling thread.
     *
     * Returns:
     *  The thread object representing the calling thread.  The result of
     *  deleting this object is undefined.
     */
    static Thread getThis()
    {
        version( Win32 )
        {
            // NOTE: this function may not be called until the module
            //       has been initialized.  this should not be an issue
            //       given the current implementation.
            return cast(Thread) TlsGetValue( sm_this );
        }
        else version( Posix )
        {
            // NOTE: this function may not be called until the module
            //       has been initialized.  this should not be an issue
            //       given the current implementation.  there is no
            //       precondition to test this because I have no idea
            //       whether zero is a legal value for an initialized
            //       pthread_key_t
            return cast(Thread) pthread_getspecific( sm_this );
        }
    }


    /**
     * This function is not intended to be used by the garbage collector,
     * so memory allocation is allowed.
     *
     * Returns:
     *  An array containing references to all threads currently being
     *  tracked by the system.  The result of
     *  deleting any contained objects is undefined.
     */
    static Thread[] getAll()
    {
        synchronized( slock )
        {
            size_t   pos = 0;
            Thread[] buf = new Thread[sm_len];

            foreach( Thread t; Thread )
            {
                buf[pos++] = t;
            }
            return buf;
        }
    }


    /**
     * Operates on all threads currently tracked by the system.
     */
    static int opApply( int delegate( inout Thread ) dg )
    {
        synchronized( slock )
        {
            int ret = 0;

            for( Thread t = sm_all; t; t = t.m_next )
            {
                ret = dg( t );
                if( ret )
                    break;
            }
            return ret;
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Global Debug Actions
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Calls suspend on all running threads.
     */
    static void suspendAll()
    {
        foreach( Thread t; Thread )
        {
            if( t.isRunning() )
                t.suspend();
            else
                remove( t );
        }
    }


    /**
     * Calls resume on all running threads.
     */
    static void resumeAll()
    {
        foreach( Thread t; Thread )
        {
            if( t.isRunning() )
                t.resume();
            else
                remove( t );
        }
    }


    ////////////////////////////////////////////////////////////////////////////
    // Local Storage Actions
    ////////////////////////////////////////////////////////////////////////////


    /**
     * Indicates the number of local storage pointers available at program
     * startup.  It is recommended that this number be at least 64.
     */
    const uint LOCAL_MAX = 64;


    /**
     * Reserves a local storage pointer for use and initializes this
     * location to null for all running threads.
     *
     * Returns:
     *  A key representing the array offset of this memory location.
     */
    static uint createLocal()
    {
        synchronized( slock )
        {
            foreach( uint key, inout ubyte set; sm_local )
            {
                if( !set )
                {
                    foreach( Thread t; sm_all )
                    {
                        t.m_local[key] = null;
                    }
                    set = true;
                    return key;
                }
            }
            throw new ThreadError( "No more local storage slots available" );
        }
    }


    /**
     * Marks the supplied key as available and sets the associated location
     * to null for all running threads.  It is assumed that any key passed
     * to this function is valid.  The result of calling this function for
     * a key which is still in use is undefined.
     *
     * Params:
     *  key = The key to delete.
     */
    static void deleteLocal( uint key )
    {
        synchronized( slock )
        {
            sm_local[key] = false;
            foreach( Thread t; sm_all )
            {
                t.m_local[key] = null;
            }
        }
    }


    /**
     * Gets the data associated with the supplied key value.  It is assumed
     * that any key passed to this function is valid.
     *
     * Params:
     *  key = The location which holds the desired data.
     *
     * Returns:
     *  The data associated with the supplied key.
     */
    static void* getLocal( uint key )
    {
        return getThis().m_local[key];
    }


    /**
     * Stores the supplied value in the specified location.  It is assumed
     * that any key passed to this function is valid.
     *
     * Params:
     *  key = The location to store the supplied data.
     *  val = The data to store.
     *
     * Returns:
     *  A copy of the data which has just been stored.
     */
    static void* setLocal( uint key, void* val )
    {
        return getThis().m_local[key] = val;
    }


protected:
    /**
     * This is the entry point for the newly invoked thread.  Default
     * behavior is to invoke the function or delegate passed on object
     * construction.  This function may be overridden to create custom
     * thread objects via subclassing.
     */
    void run()
    {
        switch( m_call )
        {
        case Call.FN:
            m_fn();
            break;
        case Call.DG:
            m_dg();
            break;
        default:
            break;
        }
    }


private:
    static Object slock()
    {
        return Thread.classinfo;
    }


    static void add( Thread t )
    in
    {
        assert( !t.m_next && !t.m_prev );
    }
    body
    {
        synchronized( slock )
        {
            if( sm_all )
            {
                t.m_next = sm_all;
                sm_all.m_prev = t;
            }
            sm_all = t;
            ++sm_len;
        }
    }


    static void remove( Thread t )
    in
    {
        assert( !t.m_next || !t.m_prev );
    }
    body
    {
        synchronized( slock )
        {
            if( t.m_prev )
                t.m_prev.m_next = t.m_next;
            if( t.m_next )
                t.m_next.m_prev = t.m_prev;
            if( sm_all == t )
                sm_all = t.m_next;
            --sm_len;
        }
        // NOTE: Don't null out t.m_next or t.m_prev because opApply currently
        //       currently follows t.m_next after removing a node.  This could
        //       be easily addressed by simply returning the next node from
        //       this function, however, a thread should never be re-added to
        //       the list anyway and having m_next and m_prev be non-null is a
        //       good way to ensure that.
    }


    enum Call
    {
        NO,
        FN,
        DG
    }


    version( Win32 )
    {
        alias uint TLSKey;
        alias uint ThreadAddr;
    }
    else version( Posix )
    {
        alias pthread_key_t TLSKey;
        alias pthread_t     ThreadAddr;
    }


    static ubyte[LOCAL_MAX] sm_local;
    static TLSKey           sm_this;

    static Thread           sm_all;
    static size_t           sm_len;

    Thread              m_prev;
    Thread              m_next;

    version( Win32 )
    {
        HANDLE          m_hndl;
        uint[8]         m_reg; // edi,esi,ebp,esp,ebx,edx,ecx,eax
    }
    else version( Posix )
    {
        void*           m_freg,
                        m_lreg;

        bool            m_isRunning;
        sem_t           m_suspended;
    }
    void*[LOCAL_MAX]    m_local;
    ThreadAddr          m_addr;
    void*               m_bstack,
                        m_tstack;
    int                 m_scount;
    Call                m_call;
    char[]              m_name;
    union
    {
        void function() m_fn;
        void delegate() m_dg;
    }
}


////////////////////////////////////////////////////////////////////////////////
// GC Support Routines
////////////////////////////////////////////////////////////////////////////////


/**
 * Adds a thread instance to the internal thread tracking mechanism
 * for tracking the main application thread.  Note that any
 * functionality related to garbage collection must operate before
 * this ctor or the module ctor has been called, and after the
 * respective dtors have been called.
 */
extern (C) void thread_init() //static this()
{
    version( Win32 )
    {
        Thread.sm_this = TlsAlloc();
        assert( Thread.sm_this != TLS_OUT_OF_INDEXES );

        Thread main   = new Thread();
        main.m_addr   = GetCurrentThreadId();
        main.m_hndl   = GetCurrentThreadHandle();
        main.m_bstack = getStackBottom();
        main.m_tstack = main.m_bstack;
        TlsSetValue( Thread.sm_this, main );
    }
    else version( Posix )
    {
        int         status;
        sigaction_t sigusr1,
                    sigusr2;

        sigusr1.sa_flags   = 0;
        sigusr1.sa_handler = &suspendHandler;
        sigemptyset( &sigusr1.sa_mask );

        sigusr2.sa_flags   = 0;
        sigusr2.sa_handler = &resumeHandler;
        sigemptyset( &sigusr2.sa_mask );

        status = sigaction( SIGUSR1, &sigusr1, null );
        assert( status == 0 );

        status = sigaction( SIGUSR2, &sigusr2, null );
        assert( status == 0 );

        status = pthread_key_create( &Thread.sm_this, null );
        assert( status == 0 );

        Thread main   = new Thread();
        main.m_addr   = pthread_self();
        main.m_bstack = getStackBottom();
        main.m_tstack = main.m_bstack;

        status = pthread_setspecific( Thread.sm_this, main );
        assert( status == 0 );
    }

    Thread.add( main );
}


/**
 * Cleans up references to any threads that are no longer running.
 */
static ~this()
{
    foreach( Thread t; Thread )
    {
        if( !t.isRunning() )
            Thread.remove( t );
    }
}


// Used for multiThreaded() below
private bool multiThreadedFlag = false;


/**
 * This function is used to determine whether the the process is
 * multi-threaded.  Optimizations may only be performed on this
 * value if the programmer can guarantee that no path from the
 * enclosed code will start a thread.
 *
 * Returns:
 *  True if Thread.start() has been called in this process.
 */
extern (C) bool thread_needLock()
{
    return multiThreadedFlag;
}


/**
 * Identical to Thread.suspendAll().
 */
extern (C) void thread_suspendAll()
{
    Thread.suspendAll();
}


/**
 * Identical to Thread.resumeAll().
 */
extern (C) void thread_resumeAll()
{
    Thread.resumeAll();
}


private alias void delegate( void*, void* ) scanAllThreadsFn;

private import std.c.stdio;

/**
 * The main entry point for garbage collection.  The supplied delegate
 * will be passed ranges representing both stack and register values.
 *
 * Params:
 *  fn          = The scanner function.  It should scan from p1 through p2 - 1.
 *  curStackTop = An optional pointer to the top of the calling thread's stack.
 */
extern (C) void thread_scanAll( scanAllThreadsFn fn, void* curStackTop = null )
{
    if( curStackTop )
    {
        if( Thread me = Thread.getThis() )
            me.m_tstack = curStackTop;
    }
    foreach( Thread t; Thread )
    {
        version( StackGrowsDown )
        {
            // NOTE: We can't index past the bottom of the stack
            //       so don't do the "+1" for StackGrowsDown.
            if( t.stackTop && t.stackTop < t.stackBottom )
                fn( t.stackTop, t.stackBottom );
        }
        else
        {
            if( t.stackBottom && t.stackBottom < t.stackTop )
                fn( t.stackBottom, t.stackTop + 1 );
        }
        if( t.regFirst && t.regFirst < t.regLast )
            fn( t.regFirst, t.regLast + 1 );
    }
}


////////////////////////////////////////////////////////////////////////////////
// ThreadGroup
////////////////////////////////////////////////////////////////////////////////


/**
 * This class is intended to simplify certain common programming techniques.
 */
class ThreadGroup
{
    /**
     * Creates and starts a new Thread object that executes fn and
     * adds it to the list of tracked threads.
     *
     * Params:
     *  fn = The thread function.
     *
     * Returns:
     *  A reference to the newly created thread.
     */
    Thread create( void function() fn )
    {
        Thread t = new Thread( fn );

        t.start();
        synchronized
        {
            m_all[t] = t;
        }
    }


    /**
     * Creates and stats a new Thread object that executes dg and
     * adds it to the list of tracked threads.
     *
     * Params:
     *  dg = The thread function.
     *
     * Returns:
     *  A reference to the newly created thread.
     */
    Thread create( void delegate() dg )
    {
        Thread t = new Thread( dg );

        t.start();
        synchronized
        {
            m_all[t] = t;
        }
    }


    /**
     * Add t to the list of tracked threads if it is not already being
     * tracked.
     *
     * Params:
     *  t - The thread to add.
     *
     * In:
     *  assert( t );
     */
    void add( Thread t )
    in
    {
        assert( t );
    }
    body
    {
        synchronized
        {
            m_all[t] = t;
        }
    }


    /**
     * Removes t from the list of tracked threads.  No operation will be
     * performed if t is not currently being tracked by this object.
     *
     * Params:
     *  t - The thread to remove.
     *
     * In:
     *  assert( t );
     */
    void remove( Thread t )
    in
    {
        assert( t );
    }
    body
    {
        synchronized
        {
            m_all.remove( t );
        }
    }


    /**
     * Operates on all threads currently tracked by this object.
     */
    int opApply( int delegate( inout Thread ) dg )
    {
        synchronized
        {
            int ret = 0;

            foreach( Thread t; m_all )
            {
                ret = dg( t );
                if( ret )
                    break;
            }
            return ret;
        }
    }


    /**
     * Iteratively joins all tracked threads.  This function
     * will block add, remove, and opApply until it completes.
     */
    void joinAll()
    {
        synchronized
        {
            foreach( Thread t; m_all )
            {
                t.join();
            }
        }
    }


private:
    Thread[Thread]  m_all;
}